//
// C++ Interface: control
//
// Description: 
//
//
// Author: Szymon Acedanski <accek@podlesie.net>, (C) 2004
// Copyright (C) 2005, 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef CONTROL_H
#define CONTROL_H

#include <stdio.h>

#include <qobject.h>
#include <qcolor.h>
#include <qfont.h>

#include "timestampedlist.h"
#include "consts.h"
#include "ar3cmonitor.h"
#include "datastore.h"
#include "offstore.h"
#include "onstore.h"
#include "niceformat.h"

class QTimer;

class BinaryObject : public QObject {
	Q_OBJECT
public:
	void EmitChange()
	{
		emit change();
	}
	
signals:
	void change();
};
	
enum cursor_mode {
	CURSOR_MODE_LRM,
	CURSOR_MODE_RLM,
	CURSOR_MODE_MLR,
};

class ControlObject : public QObject {
	Q_OBJECT
private:
	/* old access method, will be removed */
	AnalogControl *Analog;	/* single object */
	BinaryControl *Binary;	/* array */

	BinaryObject *BObjects;

	unsigned int nBinaries;
	OnlineStore Online;
	OfflineStore Offline;
	DataStore *active;	/* active store, 2.3+ */

	bool OfflineMode;
	bool OfflineDownloading;
	bool OfflineDrop;

	int layer;		/* layer id */

	struct min_req *fm_req, *lm_req;
	unsigned int mreq_count;
	unsigned long long offline_time;
	unsigned long long offline_start_ts;
	unsigned long offline_reqs;
	unsigned long lavg;
	unsigned long prev_used;

	unsigned long long TimeStamp;
	unsigned long long TimeRange;
	bool HaveData;
	bool prefetch_enabled;
	unsigned long SetupHash;
	bool ReviewMode;
	unsigned long long VGridStep;
	QColor VGridColor;
	QColor HGridColor;
	QTimer *DataTimer;
	QTimer *UpdateTimer;
	QTimer *AdjustCursorsTimer;
	unsigned long UpdateInterval;
	QFont SmallFont;
	QFont TimescaleFont;
	QFont ValueFont;
	QFont HugeFont;
	int maxBinUsed;
	unsigned long long gcollector_time;
	unsigned long gcollector_iter;
	
	prec_t prec;

	unsigned long long SyfStart;
	float SyfCurrent;
	float SyfTotal;

	enum cursor_mode cursor_mode;

	void disable_prefetch(void)
	{
		prefetch_enabled = false;
	}
	
	void enable_prefetch(void)
	{
		prefetch_enabled = true;
	}

	int shrink_online_mem(void);
	int shrink_offline_mem(int force);
	void remove_offline_R(void);
	void remove_offline_L(void);

	void newTimeStamp(unsigned long long ts);
	void autoAdjustVGrid();
	void adjustCursors();
	void downloadL();
	void downloadR();
	int __downloadOfflineData(int prefetch);
	void downloadOfflineData(int prefetch);
	void pleaseUpdate();
	void OfflinePrepareInsertIt(unsigned long id);

	BinaryControl *onlineBinary(unsigned int id);
	BinaryControl *offlineBinary(unsigned int id);
	void __garbageCollector();
	void remove_data(unsigned long long margin);

private slots:
	void garbageCollector_stats();
	void doAdjustCursors();
	void watchdog();

public:
	ControlObject(int nbinaries);
	~ControlObject();

	unsigned long mem_used()
	{
		return Offline.get_used() + Online.get_used();
	}
	
	unsigned long mem_used_all()
	{
		return Offline.get_used_all() + Online.get_used();
	}

	AnalogControl *analog()
	{
		return Analog;
	}

	BinaryControl *binary(unsigned int id)
	{
		if (id >= nBinaries) {
			qWarning("ControlObject: tried to access binary %d, "
					"while nBinaries=%d\n", id, nBinaries);
			return NULL;
		}
		return Binary + id;
	}


	BinaryObject *bobject(unsigned int id)
	{
		if (id >= nBinaries) {
			qWarning("ControlObject: tried to access bobject %d, "
					"while nBinaries=%d\n", id, nBinaries);
			return NULL;
		}
		return BObjects + id;
	}

	unsigned long long timeStamp()
	{
		return TimeStamp;
	}
	
	unsigned long long cursorStamp()
	{
		return active->get_cursor_stamp();
	}
	
	void setCursorStamp(unsigned long long stamp)
	{
		Online.set_cursor_stamp(stamp);
		Offline.set_cursor_stamp(stamp);
	}
	
	unsigned long long timeRange()
	{
		return TimeRange;
	}
	
	unsigned long long vgridStep()
	{
		return VGridStep;
	}
	
	QColor vgridColor()
	{
		return VGridColor;
	}
	
	QColor hgridColor()
	{
		return HGridColor;
	}

	int newARecord(unsigned long long ts, ARecord *r);
	void newBRecord(unsigned long long ts, int id, BState state);
	int newOffARecord(unsigned long long ts, ARecord *r);
	void newOffBRecord(unsigned long long ts, int id, BState state);
	void beginUndefinedOffline(unsigned long long ts, unsigned int flags);

	int get_layer()
	{
		return layer;
	}
	void set_layer(int layer);
	
	void garbageCollector();

	void newSyf(unsigned long long start, float current, float total);
	unsigned long long syfStart()
	{
		return SyfStart;
	}
	
	float syfCurrent()
	{
		return SyfCurrent;
	}
	
	float syfTotal()
	{
		return SyfTotal;
	}

	unsigned long long beginUndefinedRegion(unsigned long long ts);
	void ping();
	bool haveData()
	{
		return ReviewMode ? true : HaveData;
	}
	
	bool realHaveData()
	{
		return HaveData;
	}
	
	bool offlineDownloading()
	{
		return OfflineDownloading;
	}
	
	void clearData();
	void clearOfflineData();
	void clearOnlineData();
	void hashSetupData(unsigned char *buf, long len);
	void setAlwaysDefined(int id);
	void clearAlwaysDefined(int id);
	void setBinaryUsed(int id);

	bool reviewMode()
	{
		return ReviewMode;
	}
	
	bool offlineMode()
	{
		return OfflineMode;
	}

	void advanceCursor(long long amount);
	void setCursor(unsigned long long ts);
	void setTimeRange(unsigned long long range);

	void nowCanDownload();
	void end_of_data();

	const QFont &smallFont()
	{
		return SmallFont;
	}
	
	const QFont &tsFont()
	{
		return TimescaleFont;
	}
	
	const QFont &valueFont()
	{
		return ValueFont;
	}

	const QFont &hugeFont()
	{
		return HugeFont;
	}
	
	void setValueFontSize(int size);
	void setTimescaleFontSize(int size);
	void setFontSize(int size);
	void setSmallFontSize(int size);

	void set_prec(prec_t p_prec);
	prec_t get_prec(void);

	void connected();
	
	enum cursor_mode cursor_mode_get()
	{
		return cursor_mode;
	}
	
	void cursor_mode_set(enum cursor_mode mode)
	{
		cursor_mode = mode;
	}

	int penDriveConnected();

	void reset();

public slots:
	void noData();
	void reallyConnected();

	void enterReviewMode();
	void enterOfflineMode();
	void enterNormalMode();
	void enterOnlineMode();

signals:
	void sigNewTimestamp();
	void sigHaveDataChanged();
	void sigPleaseUpdate();
	void sigUpdateTimer();
	void sigModeChanged();
	void sigDownloadingOrNot();
	void sigSyfUpdate();
};

extern ControlObject *CO;

#endif
