/*
 * ar3cmonitor - datastore.h
 *
 * 	Data Store
 * 
 * Copyright (C) 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef DATASTORE_H_INCLUDED
#define DATASTORE_H_INCLUDED

#include "timestampedlist.h"
#include <errno.h>

class ARecord : public Item {
public:
	ARecord(Item *Next, Item *Prev, unsigned long long TimeStamp)
		: Item(Next, Prev, TimeStamp)
	{
	}
	
	virtual bool defined() = 0;
	virtual bool has(unsigned int i) = 0;
	virtual short get(unsigned int i) = 0;
	virtual int analog_get(unsigned int i, int *value) = 0;

	virtual ~ARecord()
	{
	}
};

class AFullRecord : public ARecord {
public:
	unsigned int count;
	short *data;

	AFullRecord(unsigned int n, Item *Next = NULL, Item *Prev = NULL,
			unsigned long long TimeStamp = 0)
		: ARecord(Next, Prev, TimeStamp), count(n)
	{
		data = (short *) mon_alloc(n * sizeof(short));
		Q_ASSERT(data);
	}
	
	~AFullRecord()
	{
		mon_free(data);
	}

	size_t get_size()
	{
		return monalloc_tsize(count * sizeof(short)) + sizeof(*this);
	}

	bool defined()
	{
		return true;
	}
	
	bool has(unsigned int i)
	{
		return i < count;
	}
	
	short get(unsigned int i)
	{
		return data[i];
	}

	int analog_get(unsigned int i, int *value)
	{
		*value = data[i];
		return 0;
	}
};

class ASingleRecord : public ARecord {
public:
	unsigned int number;
	short data;

	ASingleRecord(unsigned int n, short v, Item *Next = NULL,
			Item *Prev = NULL, unsigned long long TimeStamp = 0)
		: ARecord(Next, Prev, TimeStamp), number(n), data(v)
	{
	}
	
	~ASingleRecord()
	{
	}
	
	size_t get_size()
	{
		return sizeof(*this);
	}
	
	bool defined()
	{
		return true;
	}
	
	bool has(unsigned int i)
	{
		return i == number;
	}
	
	short get(unsigned int)
	{
		return data;
	}

	int analog_get(unsigned int i, int *value)
	{
		(void) i;
		*value = data;
		return 0;
	}
};

class AUndefinedRecord : public ARecord {
public:
	AUndefinedRecord(Item *Next = NULL, Item *Prev = NULL,
			unsigned long long TimeStamp = 0)
		: ARecord(Next, Prev, TimeStamp)
	{
	}
	
	~AUndefinedRecord()
	{
	}

	size_t get_size()
	{
		return sizeof(*this);
	}
	
	bool defined() 
	{
		return false;
	}
	
	bool has(unsigned int)
	{
		return true;
	}
	
	short get(unsigned int)
	{
		Q_ASSERT(false);
		return 0;
	}

	int analog_get(unsigned int, int *)
	{
		return -ERANGE;
	}
};

enum BState {
	Undefined,
	High,
	Low
};

class BRecord : public Item {
public:
	BState state;

	BRecord(Item *Next = NULL, Item *Prev = NULL,
			unsigned long long TimeStamp = 0)
		: Item(Next, Prev, TimeStamp)
	{
	}
	
	~BRecord()
	{		
	}
	
	size_t get_size()
	{
		return sizeof(*this);
	}
};

class AList : public TimestampedList < ARecord > {
public:
	class AIterator : public TimestampedList < ARecord >::Iterator {
	protected:
		int number;

	public:
		explicit AIterator(const Iterator &i, int num)
			: TimestampedList < ARecord >::Iterator(i),
			number(num)
		{
		}
		
		bool isPositioned()
		{
			return !valid() || data()->has(number);
		}

		AIterator &operator ++()
		{
			TimestampedList < ARecord >::Iterator::operator ++();
			positionRight();
			return *this;
		}

		void positionRight()
		{
			while (!isPositioned())
				Ptr = Ptr->Next;
		}

		AIterator &operator --()
		{
			TimestampedList < ARecord >::Iterator::operator --();
			positionLeft();
			return *this;
		}

		void positionLeft()
		{
			while (!isPositioned())
				Ptr = Ptr->Prev;
		}

		short get()
		{
			return data()->get(number);
		}
	};
};

static inline int analog_get(AList::AIterator &it, unsigned int number,
		int *value)
{
	return it.data()->analog_get(number, value);
}

struct min_req {
	unsigned long id;
	struct min_req *next;
};

class AnalogControl {
public:
	AList Data;
	AList::Iterator Cursor;
	AList::Iterator InsertIt;

	AnalogControl()
		: Cursor(Data.it()), InsertIt(Data.it())
	{
	}

	void clear()
	{
		Data.clear();
		InsertIt = Cursor = Data.it();
	}
};

class BinaryControl {
public:
	TimestampedList < BRecord > Data;
	TimestampedList < BRecord >::Iterator Cursor;
	TimestampedList < BRecord >::Iterator InsertIt;
	bool AlwaysDefined;
	bool Used;

	BinaryControl()
		: Cursor(Data.it()), InsertIt(Data.it()), AlwaysDefined(false),
		Used(false)
	{
	}
	
	void clear()
	{
		Data.clear();
		InsertIt = Cursor = Data.it();
	}
};

class DataStore {
private:
	const char *name;
	AnalogControl *Analog;
	BinaryControl *Binary;
	unsigned int nBinaries;
	unsigned long used;	/* used memory */
	unsigned long records;
	unsigned long long CursorStamp;
	
	void update_ts(unsigned long long ts);
	void error(const char *fmt, ...);
	
	void __newBRecord(unsigned long long ts, int id, BState state);
	
public:
	unsigned long long beg_ts;
	unsigned long long end_ts;
	
	DataStore(unsigned int binaries, const char *_name = NULL);

	unsigned long get_records()
	{
		return records;
	}

	void stat_new_record(size_t size)
	{
		records++;
		used += size;
	}

	unsigned long long get_cursor_stamp()
	{
		return CursorStamp;
	}

	void set_cursor_stamp(unsigned long long stamp)
	{
		CursorStamp = stamp;
	}

	void stat_delete_record(size_t size)
	{
		if (!records)
			BUG();
		records--;
		used -= size;
	}

	unsigned long get_used()
	{
		return used;
	}

	BinaryControl *get_binary(unsigned int id)
	{
		if (id >= nBinaries) {
			error("tried to access ofl binary %d, "
				"while nBinaries=%d\n", id, nBinaries);
			return NULL;
		}
		return Binary + id;
	}

	unsigned int get_binaries()
	{
		return nBinaries;
	}

	virtual int shrink(int prio);
	
	virtual void insert_ARecord(unsigned long long ts, ARecord *r);
	virtual void insert_BRecord(unsigned long long ts, int id, BRecord *r);

	virtual void newBRecord(unsigned long long ts, int id, BState state);
	virtual void newARecord(unsigned long long ts, ARecord *r);

	virtual void garbageCollector(void)
	{
	}

	void seek_eof();
	void rewind();
	void cursors_eof();

	int left_join(DataStore &a);
	int right_join(DataStore &a);
	int join(DataStore &a);

	bool empty();
	void clear();

	/* helpers for compatibility with ar3cmontor pre 2.3 - to be removed */
	AnalogControl *get_acontrol()
	{
		return Analog;
	}
	
	BinaryControl *get_bcontrol()
	{
		return Binary;
	}
	
	virtual ~DataStore();
};

#endif
