/*
 * ar3cmonitor - offstore.cpp
 *
 * 	Offline Data Store
 * 
 * Copyright (C) 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ar3cmonitor.h"
#include "datastore.h"
#include "offstore.h"
#include "control.h"
#include "commobj.h"
#include <stdio.h>
#include "analogcache.h"

#include <qobject.h>

#define OFFLINE_DEBUG	1

#ifdef OFFLINE_DEBUG
#define off_debug(fmt...)	qWarning(fmt)
#else
#define off_debug(fmt...)
#endif
	
extern unsigned int mreq_limit;

#ifdef OFFLINE_STATS

extern unsigned long long off_reqs, off_time;
static unsigned long long prev_time = 0;

#ifdef HAVE_SYS_TIME_H 

#include <sys/time.h>

static unsigned long long time_us(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return tv.tv_sec * 1000000ULL + tv.tv_usec;
}

#else

static unsigned long long time_us(void)
{
	return time(NULL) * 1000000ULL;
}

#endif
#endif

OfflineBuffer::OfflineBuffer(int binaries, unsigned long min, const char *name)
	: DataStore(binaries, name), minute(min), completed(0)
{
}

void OfflineBuffer::garbageCollector()
{
	CO->garbageCollector();
}
	
int OfflineBuffer::shrink(int)
{
	return 0;
}

OfflineBuffer::~OfflineBuffer()
{
	off_debug("OfflineBuffer Done");
}

OfflineBufferList::OfflineBufferList()
	: Sentinel(), count(0)
{
}

unsigned long OfflineBufferList::get_used_all()
{
	unsigned long ret = 0;
	OfflineBuffer *b;

	for (b = Sentinel.next; b != &Sentinel; b = b->next)
		ret += b->get_used();
	return ret;
}

OfflineBufferList::~OfflineBufferList()
{
}

OfflineStore::OfflineStore(unsigned int binaries, const char *_name)
	: DataStore(binaries, _name), MLMin(0), prefetch(1)
{
	req_limit = mreq_limit;

	do_resync = 0;
	t_counter = 0;
	wait_wdt = 0;

	BUG_ON(!empty());
}
	
void OfflineStore::buffer_finalize(OfflineBuffer *b)
{
	off_debug("finalizing %Ld (%02Ld:%02Ld) %p (%lu bytes, %lu records)",
			TS_MIN(beg_ts), TS_MIN(beg_ts) / 60,
			TS_MIN(beg_ts) % 60,
			b, b->get_used(),
			b->get_records());
	if (CO->offlineMode())
		ACC->invalidate_range_all(b->beg_ts, b->end_ts);
	BUG_ON(!MLMin && !empty());
	if (join(*b)) {
		qWarning("offline: join error");
	}
	if (!MLMin) {
		MLMin = MRMin = TS_MIN(beg_ts);
		BUG_ON(TS_MIN(end_ts) != MLMin);
	}
	t_counter = 0;
}

void OfflineStore::request_abort(void)
{
	wait_wdt = 1;
	clear_reqs();
	/* offline watchdog will try to download new data */
	t_counter = 0;
}
	
void OfflineStore::newARecord(unsigned long long ts, ARecord *r)
{
	OfflineBuffer *c = buffer_get_current();

	if (c == NULL)
		return;
	
	if (likely(TS_MIN(ts) == c->minute))
		c->newARecord(ts, r);
	else {
		qWarning("offline: analog with wrong timestamp. Aborting...");
		request_abort();
	}
	t_counter = 0;
}

void OfflineStore::newBRecord(unsigned long long ts, int id, BState state)
{
	OfflineBuffer *c = buffer_get_current();
	
	if (c == NULL)
		return;

	if (likely(TS_MIN(ts) == c->minute))
		c->newBRecord(ts, id, state);
	else {
		qWarning("offline: binary with wrong timestamp. Aborting...");
		request_abort();
	}
	t_counter = 0;
}

void OfflineStore::beginUndefined(unsigned long long ts, unsigned int)
{
	OfflineBuffer *c = buffer_get_current();
	ARecord *a;
	unsigned int i;
	
	if (c == NULL)
		return;

	a = new AUndefinedRecord();
	c->newARecord(ts, a);

	for (i = 0; i < get_binaries(); i++)
		c->newBRecord(ts, i, Undefined);
	t_counter = 0;
}

unsigned int OfflineStore::calc_prefetch(unsigned int base, unsigned int ratio,
		unsigned long long range)
{
	return base + ratio * TS_SEC(range) / 60000;
}
	
void OfflineStore::download_left()
{
	OfflineBuffer *b;

	LMin--;

	b = new OfflineBuffer(get_binaries(), LMin, "offline left buffer");
	Q_CHECK_PTR(b);
	requests.push(b);
	left.push(b);
#ifdef OFFLINE_STATS
	if (!prev_time)
		prev_time = time_us();
#endif
	TheCommObject->requestMinute(LMin);
	off_debug("requesting minute %ld (%02ld:%02ld).", LMin, LMin / 60,
			LMin % 60);
}

void OfflineStore::download_right()
{
	OfflineBuffer *b;

	RMin++;	
	b = new OfflineBuffer(get_binaries(), RMin, "offline right buffer");
	Q_CHECK_PTR(b);
	requests.push(b);
	right.push(b);
#ifdef OFFLINE_STATS
	if (!prev_time)
		prev_time = time_us();
#endif
	TheCommObject->requestMinute(RMin);
	off_debug("requesting minute %ld (%02ld:%02ld).", RMin, RMin / 60,
			RMin % 60);
}

void OfflineStore::try_merge_left()
{
	OfflineBuffer *b;

	while (((b = left.get_head()) != NULL) && b->completed) {
		if (b->minute < MLMin) {
			MLMin = b->minute;
			BUG_ON(!MLMin);
		}
		buffer_finalize(b);
		left.pop();
#ifdef OFFLINE_STATS
		unsigned long long t = time_us();
		/* I'm ignoring a race condition here */
		off_reqs++;
		off_time += t - prev_time;
		prev_time = t;
		if (!(requests.get_count()))
			prev_time = 0;
#endif
		delete b;
	}
	BUG_ON(!MLMin && !empty());
}

void OfflineStore::try_merge_right()
{
	OfflineBuffer *b;

	while (((b = right.get_head()) != NULL) && b->completed) {
		if (b->minute > MRMin)
			MRMin = b->minute;
		buffer_finalize(b);
		right.pop();
		delete b;
	}
	
	BUG_ON(!MLMin && !empty());
}

void OfflineStore::try_merge()
{
	try_merge_left();
	try_merge_right();
	
	BUG_ON(!MLMin && !empty());
}
	
void OfflineStore::finish_minute()
{
	OfflineBuffer *c = buffer_get_current();

	if (c == NULL) {
		qWarning("warning: finish_minute without request.");
		return;
	}

	c->completed = 1;
	requests.unlink(c);
	try_merge();
	
	BUG_ON(!MLMin && !empty());
}

int OfflineStore::__download()
{
	unsigned long r_real, r_max, r, l;

	if (!buffer_available()) {
		off_debug("offline: buffer not available.");
		return 0;
	}

	r_real = TS_MIN(CO->cursorStamp());

	/*
	 * FIXME: We currently limit downloading of future data just on
	 * current active layer selected by "active layer" option. However,
	 * we may also use the same layer using direct layer identifier
	 * and we should also use such limit.
	 */
	r_max = ULONG_MAX;
	if (CO->get_layer() == -1)
		r_max = TS_MIN(CO->timeStamp()) - 1;

	r = min(r_real, r_max);
	l = TS_MIN(CO->cursorStamp() - CO->timeRange());
	
	off_debug("LMin = %lu, RMin = %lu, MLMin = %lu, MRMin = %lu, l = %lu, "
			"r = %lu",
			LMin, RMin, MLMin, MRMin, l, r);
	if (empty() && requests.empty()) {
		off_debug("offline: right");
		MLMin = LMin = r + 1;
		BUG_ON(!MLMin);
		MRMin = RMin = r;
		download_left();
		return 1;
	}

	if (RMin < l) {
		if (min(l, r) - RMin > AR3CMON_OFFLINE_MAX_DELAY) {
			/* trzeba poczekac na wykonanie obecnych
			 * rzadan */
			if (!requests.empty()) {
				off_debug("offline: waiting for pending reqs.");
				return 0;
			}
			off_debug("offline: clear");
			clear();
			return __download();
		}
		
		/* we scrolled to the right completly */
		if (RMin < r_max) {
			off_debug("dOff: right");
			download_right();
			BUG_ON(!MLMin && !empty());
			return 1;
		}
		BUG_ON(!MLMin && !empty());
	} else if (RMin < r) {
		off_debug("dOff: right");
		download_right();
		BUG_ON(!MLMin && !empty());
		return 1;
	}
	
	if (LMin > l) {
		if (r < LMin && LMin - r > AR3CMON_OFFLINE_MAX_DELAY) {
			/* trzeba poczekac na wykonanie obecnych
			 * rzadan */
			if (!requests.empty()) {
				off_debug("offline: waiting for pending reqs.");
				return 0;
			}
			off_debug("offline: clear");
			clear();
			return __download();
		}
		off_debug("dOff: left");
		download_left();
		BUG_ON(!MLMin && !empty());
		return 1;
	}

	if (LMin > l - calc_prefetch(AR3CMON_OFFLINE_LPREFETCH,
				AR3CMON_OFFLINE_LPREFETCH_R, CO->timeRange())) {
		off_debug("prefetch left");
		download_left();
		return 1;
	}
	
	if ((RMin < r + calc_prefetch(AR3CMON_OFFLINE_RPREFETCH,
				AR3CMON_OFFLINE_RPREFETCH_R, CO->timeRange()))
			&& (RMin < r_max)) {
		off_debug("prefetch right");
		download_right();
		return 1;
	}

	off_debug("no decision");
	BUG_ON(!MLMin && !empty());
	return 0;
}

void OfflineStore::request_lmin_high()
{	
	OfflineBuffer *b;
	b = new OfflineBuffer(get_binaries(), MLMin, "offline left buffer");
	Q_CHECK_PTR(b);
	requests.push(b);
	left.push_head(b);
	TheCommObject->requestMinute(MLMin);
	off_debug("shrink: retransmit minute %ld (%02ld:%02ld).", MLMin,
			MLMin / 60, MLMin % 60);
}

void OfflineStore::request_rmin_high()
{	
	OfflineBuffer *b;
	b = new OfflineBuffer(get_binaries(), MRMin, "offline left buffer");
	Q_CHECK_PTR(b);
	requests.push(b);
	right.push_head(b);
	TheCommObject->requestMinute(MRMin);
	off_debug("shrink: retransmit minute %ld (%02ld:%02ld).", MRMin,
			MRMin / 60, MRMin % 60);
	BUG_ON(!MLMin && !empty());
}

void OfflineStore::remove_left()
{
	unsigned int i;
	unsigned long long margin;
	AnalogControl *a = get_acontrol();
	AList *al = &a->Data;
	Item *item;

	if (LMin < MLMin)
		request_lmin_high();
	else
		LMin++;
	MLMin++;
	
	margin = MLMin * 60000000000ULL;
	if (MLMin > MRMin)
		MLMin = 0;
	
	/* We don't care about InsertIt, because we use only join() */
	while (!al->empty() && (al->head()->TimeStamp < margin)) {
		item = al->poph();
		
		if (a->Cursor == item)
			++(a->Cursor);

		stat_delete_record(monalloc_tsize(item->get_size()));
		delete item;
	}

	BUG_ON(!MLMin && !al->empty());

	for (i = 0; i < get_binaries(); i++) {
		BinaryControl *b = get_bcontrol() + i;
		TimestampedList<BRecord> *bl = &b->Data;
		
		while (!bl->empty() && (bl->head()->TimeStamp < margin)) {
			item = bl->poph();

			if (b->Cursor == item)
				++(b->Cursor);
			
			stat_delete_record(monalloc_tsize(item->get_size()));
			delete item;
		}
		BUG_ON(!MLMin && !bl->empty());
	}
	
	if (MLMin > TS_MIN(beg_ts))
		beg_ts = MIN_TS(MLMin);
	
	BUG_ON(!MLMin && !empty());
}

void OfflineStore::remove_right()
{
	unsigned int i;
	unsigned long long margin = MIN_TS(MRMin);
	AnalogControl *a = get_acontrol();
	AList *al = &a->Data;
	Item *item;
	
	while (!al->empty() && (al->tail()->TimeStamp >= margin)) {
		item = al->popt();
		
		if (a->Cursor == item)
			--(a->Cursor);

		stat_delete_record(monalloc_tsize(item->get_size()));
		delete item;
	}

	for (i = 0; i < get_binaries(); i++) {
		BinaryControl *b = get_bcontrol() + i;
		TimestampedList<BRecord> *bl = &b->Data;
	
		while (!bl->empty() && (bl->tail()->TimeStamp >= margin)) {
			item = bl->popt();
			
			if (b->Cursor == item)
				--(b->Cursor);
			
			stat_delete_record(monalloc_tsize(item->get_size()));
			delete item;
		}
	}

	if (RMin > MRMin)
		request_rmin_high();
	else
		RMin--;
	
	if (MRMin <= TS_MIN(end_ts))
		end_ts = MIN_TS(MRMin) - 1;
	MRMin--;
	
	if (MLMin > MRMin) {
		if (!MLMin && !al->empty()) {
			off_debug("MLMin: %lu MRMin: %lu LMin: %lu RMin: %lu",
					MLMin, MRMin, LMin, RMin);
			off_debug("beg: %Lu, end: %Lu", TS_MIN(beg_ts),
					TS_MIN(end_ts));
			while (1) { };
		}	
		MLMin = 0;

		BUG_ON(!empty());
	}
	for (i = 0; i < get_binaries(); i++) {
		BinaryControl *b = get_bcontrol() + i;
		TimestampedList<BRecord> *bl = &b->Data;
		BUG_ON(!MLMin && !bl->empty());
	}
	
	BUG_ON(!MLMin && !empty());
}

int OfflineStore::shrink_buffers(int)
{
	/* unsupported yet */
	return 0;
}

int OfflineStore::shrink(int prio)
{
	unsigned long r = TS_MIN(CO->cursorStamp());
	unsigned long l = TS_MIN(CO->cursorStamp() - CO->timeRange());
	long la, ra;

	if (empty())
		return 0;

	BUG_ON(MLMin == 0);

	if (!CO->offlineMode()) {
		remove_left();
		disable_prefetch();
		return 1;
	}

	la = (long) l - MLMin;
	ra = (long) MRMin - r;

	if (la >= ra) {
		if (ra > 0) {
			remove_left();
			disable_prefetch();
			BUG_ON(!MLMin && !empty());
			return 1;
		}
	} else if (la > 0) {
		remove_right();
		disable_prefetch();
		BUG_ON(!MLMin && !empty());
		return 1;
	}

	if (prio) {
		off_debug("offline: force remove");
		remove_right();
		disable_prefetch();
		BUG_ON(!MLMin && !empty());
		return 1;
	}
	
	BUG_ON(!MLMin && !empty());
	
	return 0;
}

void OfflineStore::download()
{	
	if (!CO->offlineMode())
		return;
	if (do_resync)
		return;
	while (__download());
	BUG_ON(!MLMin && !empty());
}

unsigned long OfflineStore::get_used_bufs()
{
	return requests.get_used_all();
}

unsigned long OfflineStore::get_used_all()
{
	return get_used() + requests.get_used_all();
}

void OfflineStore::garbageCollector()
{
	CO->garbageCollector();
}

void OfflineStore::clear_reqs()
{
	OfflineBuffer *b;
	unsigned int i = 0;

	left.clear();
	right.clear();

	while ((b = requests.pop()) != NULL) {
		delete b;
		i++;
	}
	if (requests.get_count())
		qWarning("request count != 0.");
	LMin = MLMin;
	RMin = MRMin;
	off_debug("removed: %d", i);
}

void OfflineStore::connected()
{
	clear_reqs();
	download();
	t_counter = 0;
}

void OfflineStore::resync()
{
	do_resync = 1;
	wait_wdt = 1;
}

void OfflineStore::watchdog()
{
	if (do_resync) {
		if (!requests.get_count()) {
			do_resync = 0;
			wait_wdt = 0;
			clear_reqs();
			t_counter = 0;
			CO->clearOfflineData();
			download();
		}
	}
	/* FIXME: fix race condition */
	if (t_counter++ > 20) {
		wait_wdt = 0;
		if (!requests.empty()) {
			qWarning("offline watchdog: timedout");
			clear_reqs();
		}
		do_resync = 0;
		download();
		t_counter = 0;
	}
}

void OfflineStore::please_update()
{
	if (!wait_wdt && requests.empty())
		download();
}

void OfflineStore::clear()
{
	unsigned int i;
	AnalogControl *a = get_acontrol();
	AList *al = &a->Data;

	off_debug("offline:clear()");
	DataStore::clear();
	MLMin = 0;
	BUG_ON(!empty());

	BUG_ON(!MLMin && !al->empty());
	for (i = 0; i < get_binaries(); i++) {
		BinaryControl *b = get_bcontrol() + i;
		TimestampedList<BRecord> *bl = &b->Data;
		BUG_ON(!MLMin && !bl->empty());
	}

	BUG_ON(!MLMin && !empty());
}

OfflineStore::~OfflineStore()
{
}
