/*
 * ar3cctrl - signature.c
 * Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#ifdef HAVE_SSL
#include <openssl/sha.h>
#endif
#include <stdlib.h>
#include <string.h>

#include <ar3cm_proto.h>
#include <signature.h>

int signature_init(struct signature *s)
{
	s->token = NULL;
	s->priv = NULL;
	return 0;
}

int signature_set_token(struct signature *s, struct token *token)
{
	s->token = token;
	return 0;
}

#ifdef HAVE_SSL
static void initialize_sha1_signature(struct signature *s)
{
	SHA1_Init(s->priv);
}
#endif

static int signature_initialize_priv(struct signature *s)
{
	if (!s->token)
		return -EINVAL;

#ifdef HAVE_SSL
	if (s->token->type == AR3CM_TOKEN_SHA1)
		initialize_sha1_signature(s);
	else
#endif
		return -EINVAL;
	return 0;
}

int signature_reset(struct signature *s)
{
	if (!s->priv) {
#ifdef HAVE_SSL
		if (s->token->type == AR3CM_TOKEN_SHA1)
			s->priv = malloc(sizeof(SHA_CTX));
		else
#endif
			return -EINVAL;
	}
	return signature_initialize_priv(s);
}

int signature_update(struct signature *s, const void *buf, size_t size)
{
#ifdef HAVE_SSL
	SHA1_Update(s->priv, buf, size);
	return 0;
#else
	return -EINVAL;
#endif
}

int signature_size(struct signature *s)
{
	if (s->token == NULL)
		return -EINVAL;

#ifdef HAVE_SSL
	if (s->token->type == AR3CM_TOKEN_SHA1)
		return SHA_DIGEST_LENGTH;
#endif
	return -EINVAL;
}

static int signature_finalize(struct signature *s)
{
	if (s->token == NULL)
		return -EINVAL;

#ifdef HAVE_SSL
	if (s->token->type == AR3CM_TOKEN_SHA1)
		SHA1_Update(s->priv, s->token->data, s->token->size);
	else
#endif
		return -EINVAL;
	return 0;
}

int signature_generate(struct signature *s, void *buf)
{
	int ret;

	ret = signature_finalize(s);
	if (ret)
		return ret;
#ifdef HAVE_SSL
	if (s->token->type == AR3CM_TOKEN_SHA1)
		SHA1_Final(buf, s->priv);
	else
#endif
		return -EINVAL;
	return 0;
}

#ifdef HAVE_SSL
static int signature_verify_sha1(struct signature *s, const void *buf,
		size_t size)
{
	unsigned char sign[SHA_DIGEST_LENGTH];

	if (size != SHA_DIGEST_LENGTH)
		return -EINVAL;

	SHA1_Final(sign, s->priv);
	return memcmp(sign, buf, SHA_DIGEST_LENGTH) ? -EINVAL : 0;
}
#endif

int signature_verify(struct signature *s, const void *buf, size_t size)
{
	int ret;

	ret = signature_finalize(s);
	if (ret)
		return ret;
#ifdef HAVE_SSL
	if (s->token->type == AR3CM_TOKEN_SHA1)
		return signature_verify_sha1(s, buf, size);
#endif
	return -EINVAL;
}

int signature_destroy(struct signature *s)
{
	if (s->priv)
		free(s->priv);
	return 0;
}
