/*
 * ar3cmonitor - tscale.h
 *
 * Copyright (C) 2011  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __TSCALE_H
#define __TSCALE_H

#include <limits.h>
#include <stdint.h>

/*
 * The scaling code use (1 << T_SCALER) ns quantum. The range must
 * be smaller than 2^BITS_PER_LONG quants.
 *
 * T_SCALER	quantum		32-bit range	64-bit range
 * 0		1 ns		4.2 s		580 years
 * 12		1 ns		4.8 hours	2 milion years
 * 16		1 ns		78 hours	38 milion years
 */
#define T_SCALER	12	/* 4.8 hours on 32-bit systems */

/* support for time ranges up to 4 hours */
#define SCALER_MIN_RANGE	(4 * 3600 * 1000000000ULL >> T_SCALER)

#if defined(__i386__) && defined(__GNUC__)
static inline uint64_t div_u64(uint64_t dividend, uint32_t divisor)
{
	union {
		uint64_t v64;
		uint32_t v32[2];
	} d = { dividend };
	uint32_t upper;

	upper = d.v32[1];
	d.v32[1] = 0;
	if (upper >= divisor) {
		d.v32[1] = upper / divisor;
		upper %= divisor;
	}
	asm ("divl %1" : "=a" (d.v32[0])
			: "rm" (divisor), "0" (d.v32[0]), "d" (upper));
	return d.v64;
}

static inline long long tscale_div(long long dividend, unsigned int divisor)
{
	if (dividend >= 0) {
		dividend = div_u64(dividend, divisor);
	} else {
		dividend = -dividend;
		dividend = div_u64(dividend, divisor);
		dividend = -dividend;
	}
	return dividend;
}
#elif (LONG_MAX >= SCALER_MIN_RANGE)
static inline long long tscale_div(long long dividend, long divisor)
{
	return dividend / divisor;
}
#elif (ULONG_MAX >= SCALER_MIN_RANGE)
static inline long long tscale_div(long long dividend, unsigned long divisor)
{
	unsigned long long x;

	if (dividend >= 0) {
		x = dividend;
		x /= divisor;
		return x;
	}
	x = -dividend;
	x /= divisor;
	return -((long long) x);
}
#else
static inline long long tscale_div(long long dividend, long long divisor)
{
	return dividend / divisor;
}
#endif

/**
 * tscale - 64-bit scaling
 * @t: number to scale
 * @t0: base
 * @w: new range
 * @range: old range
 *
 * This function performs 64-bit scaling. The output is equal to:
 *	w / range * (t - t0)
 *
 * It's assumed that:
 *	- (t - t0) never overflows, this limits effective range
 *	  to 63-bits, the range is limited to year 2262
 *
 *      - maximal timerange is equal to 4.8 hours by default
 */
static long long tscale(unsigned long long t, unsigned long long t0,
		int w, unsigned long long range)
{
	long long l = (long long) t - (long long) t0;

	l >>= T_SCALER;
	range >>= T_SCALER;
	l *= w;
	return tscale_div(l, range);
}

#endif
